//====================================================================
//Project Lynxmotion Phoenix
//
// Servo Driver: This version is setup to use the main processor to
//    drive the servos, using my hacked library ServoEx, which is based
//    on the Servo class. 
//====================================================================
#if ARDUINO>99
#include <Arduino.h> // Arduino 1.0
#else
#include <Wprogram.h> // Arduino 0022
#endif
#include "Hex_Globals.h"
#include "ServoDriver.h"

#ifndef USE_SSC32

#include <ServoEx.h>

// Add support for running on non-mega Arduino boards as well.
#ifdef __AVR__
#if not defined(UBRR1H)
    SoftwareSerial SSCSerial(cSSC_IN, cSSC_OUT);
#endif
#endif

//=============================================================================
// Global - Local to this file only...
//=============================================================================
#ifdef c4DOF
#define NUMSERVOSPERLEG 4
#else
#define NUMSERVOSPERLEG 3
#endif
ServoEx          g_aservoLegs[6*NUMSERVOSPERLEG];        // Define the servo objects...
short            g_asLegOffsets[6*NUMSERVOSPERLEG];       // Offsets per leg
cServoGroupMove  g_cSGM;
boolean g_fServosAttached;


//--------------------------------------------------------------------
// Helper function to load the servo offsets from the EEPROM
//--------------------------------------------------------------------
void LoadServosConfig(void) {

  byte *pb = (byte*)&g_asLegOffsets;
  byte bChkSum = 0;      //
  int i;

  if (EEPROM.read(0) == 6*NUMSERVOSPERLEG) {
    for (i=0; i < sizeof(g_asLegOffsets); i++) {
          *pb = EEPROM.read(i+2);
          bChkSum += *pb++;
    }
     
    // now see if the checksum matches.
    if (bChkSum == EEPROM.read(1))
      return;    // we have valid data
  }
  
  // got to here, something not right, set up 0's for servo offsets.
  for (i = 0; i < 6*NUMSERVOSPERLEG; i++) {
    g_asLegOffsets[i] = 0;
  }   

}


//--------------------------------------------------------------------
//Init
//--------------------------------------------------------------------
void ServoDriver::Init(void) {
  g_fServosAttached = false;  // remember we are not attached. Could simply ask one of our servos...
  
  // Need to read in the servo offsets... but for now just init all to 0
  for (byte bServoIndex = 0; bServoIndex < 6*NUMSERVOSPERLEG; bServoIndex++)
      g_asLegOffsets[bServoIndex] = 0;
    
}

//--------------------------------------------------------------------
//[FIsGPEnabled]
//--------------------------------------------------------------------
boolean ServoDriver::FIsGPEnabled(void)
{
  // Assume for now that we don't support GP Player
  return false;
}


//--------------------------------------------------------------------
//[GP PLAYER]
void ServoDriver::GPPlayer(void)
{
}

//--------------------------------------------------------------------
// Attach Servos...
//--------------------------------------------------------------------
void AttachServos(void) {
  byte bServoIndex;
  if (!g_fServosAttached) {
    bServoIndex = 0;
    for ( byte LegIndex=0; LegIndex < 6; LegIndex++) {
      // BUGBUG:: will probably need to add additional stuff here to get the servo offsets
      // and calculate min/max to use...
      g_aservoLegs[bServoIndex++].attach(pgm_read_byte(&cCoxaPin[LegIndex]));
      g_aservoLegs[bServoIndex++].attach(pgm_read_byte(&cFemurPin[LegIndex]));
      g_aservoLegs[bServoIndex++].attach(pgm_read_byte(&cTibiaPin[LegIndex]));
#ifdef c4DOF
      g_aservoLegs[bServoIndex++].attach(pgm_read_byte(&cTarsPin[LegIndex]));
#endif
    }
    g_fServosAttached = true;
  }
}


//--------------------------------------------------------------------
//[SERVO DRIVER Start] Updates the positions of the servos - This outputs
//         as much of the command as we can without committing it.  This
//         allows us to once the previous update was completed to quickly 
//        get the next command to start
//--------------------------------------------------------------------
#define cPwmDiv       991  //old 1059;
#define cPFConst      592  //old 650 ; 900*(1000/cPwmDiv)+cPFConst must always be 1500
                           // A PWM/deg factor of 10,09 give cPwmDiv = 991 and cPFConst = 592
                           // For a modified 5645 (to 180 deg travel): cPwmDiv = 1500 and cPFConst = 900.

void ServoDriver::UpdateServoDriver()
{
    byte   LegIndex;
    byte   bServoIndex;
    word    wCoxaSSCV;        // Coxa value in SSC units
    word    wFemurSSCV;        //
    word    wTibiaSSCV;        //
#ifdef c4DOF
    word    wTarsSSCV;        //
#endif

    // First lets make sure our servos are attached.
    AttachServos();

    g_cSGM.start();    // tell the group move system we are starting a new move
#ifdef DEBUG_GAITS
    if (g_fDebugOutput && g_fHexOn && (TravelLengthX || TravelLengthZ))
        DBGSerial << _DEC(GaitType) << " " << _DEC(GaitStep) << " " << _DEC(BalanceMode) << ":"; 
#endif
  

    bServoIndex = 0;
    for (LegIndex = 0; LegIndex <= 5; LegIndex++) {
        if (LegIndex < 3) {
            wCoxaSSCV = ((long)(-CoxaAngle1[LegIndex] +900))*1000/cPwmDiv+cPFConst;
            wFemurSSCV = ((long)(-FemurAngle1[LegIndex]+900))*1000/cPwmDiv+cPFConst;
            wTibiaSSCV = ((long)(-TibiaAngle1[LegIndex]+900))*1000/cPwmDiv+cPFConst;
#ifdef c4DOF
            wTarsSSCV = ((long)(-TarsAngle1[LegIndex]+900))*1000/cPwmDiv+cPFConst;
#endif
        } else {
            wCoxaSSCV = ((long)(CoxaAngle1[LegIndex] +900))*1000/cPwmDiv+cPFConst;
            wFemurSSCV = ((long)((long)(FemurAngle1[LegIndex]+900))*1000/cPwmDiv+cPFConst);
            wTibiaSSCV = ((long)(TibiaAngle1[LegIndex]+900))*1000/cPwmDiv+cPFConst;
#ifdef c4DOF
            wTarsSSCV = ((long)(TarsAngle1[LegIndex]+900))*1000/cPwmDiv+cPFConst;
#endif
        }

        // Now lets tell the servos their next  location...
        g_aservoLegs[bServoIndex].writeMicroseconds(wCoxaSSCV + g_asLegOffsets[bServoIndex]);
        bServoIndex++;
        g_aservoLegs[bServoIndex].writeMicroseconds(wFemurSSCV + g_asLegOffsets[bServoIndex]);
        bServoIndex++;
        g_aservoLegs[bServoIndex].writeMicroseconds(wTibiaSSCV + g_asLegOffsets[bServoIndex]);
        bServoIndex++;
#ifdef c4DOF
        g_aservoLegs[bServoIndex].writeMicroseconds(wTarsSSCV);
        bServoIndex++;
#endif
#ifdef DEBUG_GAITS
        if (g_fDebugOutput && g_fHexOn && (TravelLengthX || TravelLengthZ)) {
            DBGSerial << _DEC((short)pgm_read_byte(&cCoxaPin[LegIndex])) << "=(" << _DEC(CoxaAngle1[LegIndex]) << ")" << _DEC(wCoxaSSCV) << " ";
            DBGSerial << _DEC((short)pgm_read_byte(&cFemurPin[LegIndex])) << "=(" << _DEC(FemurAngle1[LegIndex]) << ")" << _DEC(wFemurSSCV) << " ";
            DBGSerial << _DEC((short)pgm_read_byte(&cTibiaPin[LegIndex])) << "=(" << _DEC(TibiaAngle1[LegIndex]) << ")" << _DEC(wTibiaSSCV) << " ";
        }    
#endif    
    }  
}


//--------------------------------------------------------------------
//[CommitServoDriver Updates the positions of the servos - This outputs
//         as much of the command as we can without committing it.  This
//         allows us to once the previous update was completed to quickly 
//        get the next command to start
//--------------------------------------------------------------------
void ServoDriver::CommitServoDriver()
{
    g_cSGM.commit(ServoMoveTime);

#ifdef DEBUG_GAITS
    if (g_fDebugOutput && g_fHexOn && (TravelLengthX || TravelLengthZ)) {
	DBGSerial << "T:" << _DEC(ServoMoveTime) << endl; 
    }
#endif	

    PrevServoMoveTime = ServoMoveTime;
}

//--------------------------------------------------------------------
//[FREE SERVOS] Frees all the servos
//--------------------------------------------------------------------
void ServoDriver::FreeServos(void)
{
  for (byte bServoIndex=0; bServoIndex < 6*NUMSERVOSPERLEG; bServoIndex++) {
    // BUGBUG:: will probably need to add additional stuff here to get the servo offsets
    // and calculate min/max to use...
    g_aservoLegs[bServoIndex].detach();
  }
  g_fServosAttached = false;
}


 //==============================================================================
// SSC Forwarder - used to allow things like Lynxterm to talk to the SSC-32 
// through the Arduino...  Will see if it is fast enough...
//==============================================================================
void  ServoDriver::SSCForwarder(void) 
{
    DBGSerial.println("SSC Forwarder not supported");
}


//=================================================================================================================
void AllLegServos1500() 
{
    for (byte bServoIndex=0; bServoIndex < 6*NUMSERVOSPERLEG; bServoIndex++) {
        g_aservoLegs[bServoIndex].writeMicroseconds(1500 + g_asLegOffsets[bServoIndex]);
    }
}

//==============================================================================
//	FindServoOffsets - Find the zero points for each of our servos... 
// 		Will use the new servo function to set the actual pwm rate and see
//		how well that works...
//==============================================================================
void ServoDriver::FindServoOffsets()
{
    // not clean but...
    static char *apszLegs[] = {"RR","RM","RF", "LR", "LM", "LF"};  // Leg Order
    static char *apszLJoints[] = {" Coxa", " Femur", " Tibia", " tArs"}; // which joint on the leg...
    int data;
    short sSN = 0; 			// which servo number
    boolean fNew = true;	// is this a new servo to work with?
    boolean fExit = false;	// when to exit
    int ich;
    short sOffset;
    
    if (CheckVoltage()) {
        // Voltage is low... 
        Serial.println("Low Voltage: fix or hit $ to abort");
        while (CheckVoltage()) {
            if (Serial.read() == '$')  return;
        }
    }
        
        
    // OK lets move all of the servos to their zero point.
    Serial.println("Find Servo Zeros.\n$-Exit, +- changes, *-change servo");
    Serial.println("    0-5 Chooses a leg, C-Coxa, F-Femur, T-Tibia");
    
    // don't continue here until we have a valid voltage to work with.
    AttachServos();
    AllLegServos1500();


    while(!fExit) {
        if (fNew) {
            sOffset = g_asLegOffsets[sSN];
            Serial << "Servo: " <<  apszLegs[sSN/NUMSERVOSPERLEG] << apszLJoints[sSN%NUMSERVOSPERLEG] << "(" << _DEC(sOffset) << ")" << endl; 

	    // Now lets wiggle the servo
            g_cSGM.wait(0xffffff);    // wait for any active servos to finish moving...
            g_aservoLegs[sSN].move(1500+sOffset+250, 500);

            g_cSGM.wait(0xffffff);    // wait for any active servos to finish moving...
            g_aservoLegs[sSN].move(1500+sOffset-250, 500);

            g_cSGM.wait(0xffffff);    // wait for any active servos to finish moving...
            g_aservoLegs[sSN].move(1500+sOffset, 500);
            fNew = false;
        }
	
	//get user entered data
	data = Serial.read();
	//if data received
	if (data !=-1) 	{
            if (data == '$')
		fExit = true;	// not sure how the keypad will map so give NL, CR, LF... all implies exit

	    else if ((data == '+') || (data == '-')) {
	        if (data == '+')
		    sOffset += 5;		// increment by 5us
		else
		    sOffset -= 5;		// increment by 5us

		Serial <<"    " << _DEC(sOffset) << endl; 
                g_asLegOffsets[sSN] = sOffset;
                g_aservoLegs[sSN].move(1500+sOffset, 500);
	    } else if ((data >= '0') && (data <= '5')) {
		// direct enter of which servo to change
		fNew = true;
		sSN = (sSN % NUMSERVOSPERLEG) + (data - '0')*NUMSERVOSPERLEG;
	    } else if ((data == 'c') && (data == 'C')) {
		fNew = true;
		sSN = (sSN / NUMSERVOSPERLEG) * NUMSERVOSPERLEG + 0;
	    } else if ((data == 'c') && (data == 'C')) {
		fNew = true;
		sSN = (sSN / NUMSERVOSPERLEG) * NUMSERVOSPERLEG + 1;
	    } else if ((data == 'c') && (data == 'C')) {
		// direct enter of which servo to change
		fNew = true;
		sSN = (sSN / NUMSERVOSPERLEG) * NUMSERVOSPERLEG + 2;
	    } else if (data == '*') {
	        // direct enter of which servo to change
		fNew = true;
		sSN++;
		if (sSN == 6*NUMSERVOSPERLEG) 
		    sSN = 0;	
	    }
	}
    }
    Serial.print("Find Servo exit ");
    for (sSN=0; sSN < 6*NUMSERVOSPERLEG; sSN++){
        Serial.print(" ");
        Serial.print(g_asLegOffsets[sSN], DEC);
    }

    Serial.print("\nSave Changes? Y/N: ");

    //get user entered data
    while (((data = Serial.read()) == -1) || ((data >= 10) && (data <= 15)))
	; 

    if ((data == 'Y') || (data == 'y')) {
        // Ok they asked for the data to be saved.  We will store the data with a 
        // number of servos (byte)at the start, followed by a byte for a checksum...followed by our offsets array...
        // Currently we store these values starting at EEPROM address 0. May later change...
	// 
        byte *pb = (byte*)&g_asLegOffsets;
        byte bChkSum = 0;  //
        EEPROM.write(0, 6*NUMSERVOSPERLEG);    // Ok lets write out our count of servos
	for (sSN=0; sSN < sizeof(g_asLegOffsets); sSN++) {
            EEPROM.write(sSN+2, *pb);
            bChkSum += *pb++;
	}
        // Then write out to address 1 our calculated checksum
        EEPROM.write(1, bChkSum);
    } else {
        void LoadServosConfig();
    }
    
    FreeServos();

}


#endif // USE_SSC32
